;               ************************************************
;               *  Sample LIAL program to find Mersenne primes *
;               ************************************************
;
;********************************* REMARKS ***********************************
;
; This LIAL program finds prime numbers of the form 2**P - 1.  Such a number
; can only be prime when P is also prime.  These are called Mersenne primes.
; The first few Mersenne primes have P = 2, 3, 5, 7, 13, 17, 19, 31, ... as
; their exponents.  The largest known primes of any kind are Mersenne primes.
; This is due to a relatively fast way of testing them known as the Lucas-
; Lehmer test.  The Lucas-Lehmer test (used by this program) runs as follows:
;
;           Let P be an odd prime.  Define a sequence L(K) by
;
;                1.  L(0) = 4
;                2.  L(K) = L(K-1)*L(K-1) - 2 mod 2**P - 1
;
;           Then 2**P - 1 is prime if and only if L(P-2) is zero.
;
; The program takes as input a two-byte integer in registers R0 (high) and R1
; (low).  The program output is the first odd Mersenne exponent larger than the
; input value.  The result is returned in R0 and R1.  Subsequent program
; restarts will find successively higher Mersenne exponents.  (A list of odd
; Mersenne exponents found is also saved above the 24 K point of high storage,
; and the running count of the number found is saved in R6.)
;
; The program dynamically sets the long integer size to smallest needed for
; the current test.  The largest long integer size possible here is 8K.  The
; declared storage below is three times that plus 1.5 K, since the accumulator
; and two long integers are needed during a test and since an extra 1.5 K
; (more than adequate) is used to save the list of Mersenne exponents found.
;
; The running time for a test of one exponent is roughly proportional to the
; cube of the exponent, if the LIAL interpreter uses the classical N-square
; multiplication algorithm.
;
;************************** ASSEMBLER ASSIGNMENTS ****************************
 
; First, initialize some object file header values:
 
* R0      = 0      ; high-low starting point for         \  ******************
* R1      = 1      ;   Mersenne exponent search           > * MAY BE CHANGED *
* FLAGS   = $1     ; enables user break                  /  ******************
 
* STORAGE = $6600  ; three times 8K plus 1.5 K required
* PC      = BOOT   ; initialize program counter (default also ok here)
* R6      = 0      ; Mersenne prime counter initial value (default ok)
 
; Next, assign variable names to lowest available of the 16 long integers:
 
# LVAL    = 0      ; holds Lucas-Lehmer sequence values
# LDUP    = 1      ; holds copy for square operation
 
; Following name is long integer alias (used in GETPRM when N = 2):
 
# TRY     = 0      ; used when finding next prime exponent try
 
; Following name is last long integer (used in SAVMER when N = 1.5 K):
 
# LAST    = 15     ; holds list of (up to 768) Mersenne exponents found
 
; Next, assign variable names to Z-registers that are used:
 
# HITRY   = 0      ; high-low pair holds current Mersenne
# LOTRY   = 1      ;   exponent try
# TEMP    = 2      ; temporary use
# SHIFT   = 3      ; holds bitshifts required for mod computation
# HICNT   = 4      ; high-low pair, initialized to exponent try less three
# LOCNT   = 5      ;   and used to count sequence iterations
# MERCNT  = 6      ; holds count of Mersenne exponents found
 
;****************************** PROGRAM **************************************
 
; Note: the END instruction below may be changed to NOP or deleted to have
; the program run continuously until the maximum two-byte exponent is
; reached (which will cause the program to terminate in subroutine GETPRM).
 
; If this is done, make sure that FLAGS allow user breaks or mul-div breaks
; to periodically check intermediate results.  The running time for the
; complete program on an AT speed IBM-compatible would be centuries.  However,
; all Mersenne primes with exponents less than 2000 may be determined in about
; a day on such a machine.
 
BOOT   JSR GETTRY  ; gets next prime exponent try and readies for test
       JSR PRETST  ; checks for a few small divisors first ... Z holds result
       BZN BOOT
       JSR LLTEST  ; does one Lucas-Lehmer test ... ACC holds result
       BNE BOOT
       JSR SAVMER  ; saves HITRY, LOTRY in high storage above 24 K area
       END         ; (may be changed to NOP...see comment above)
       BRN BOOT    ; restart to get next Mersenne exponent...
 
; GETTRY sets (HITRY,LOTRY) to next prime exponent try,
;        sets (HICNT,LOCNT) to try less three (latch counter for LLTEST),
;        sets long integer size (N) to minimum value required, and
;        sets SHIFT to the number of excess bits in N over try size
 
GETTRY CLZ         ; first, set long integer size to two bytes
       LNH
       LZI 2
       LNL
 
       ZER         ; next, set TRY to previous try (or initial program input)
       LDZ HITRY
       RLZ
       LDZ LOTRY
       RLZ
       RRC         ; insure odd (in case initial program input even)
       SEC
       RLC
       STA TRY
 
       JSR GETPRM  ; sets TRY to next higher odd prime
 
       LDA TRY     ; update (HITRY,LOTRY)
       RRZ
       STZ LOTRY
       RRZ
       STZ HITRY
 
       LDA TRY     ; set (HICNT,LOCNT) latch to TRY - 3
       DEC
       DEC
       DEC
       RRZ
       STZ LOCNT
       RRZ
       STZ HICNT
 
       LAI 7       ; set SHIFT to 1, 3, 5, or 7
       AND TRY
       LZI 8
       SBZ
       NEG
       RRZ
       STZ SHIFT
 
       LDA TRY     ; set long integer size to smallest that will
       LZI 8       ;   accommodate 2**TRY - 1 ... SHIFT is number
       DVZ         ;   of leading clear bits that would remain
       INC
       RRZ
       STZ TEMP
       RRZ
       LNH
       LDZ TEMP
       LNL
       RTS
 
; GETPRM increases TRY to next prime value...TRY must be odd
 
GETPRM LDA TRY     ; increment TRY by 2
       INC
       BNE DOWN
BACK   END         ; terminate program ... exponent limit reached
       BRN BACK
DOWN   INC
       STA TRY
       LZI 1
       STZ TEMP
INLP   LDZ TEMP    ; inner loop checks for divisors 3, 5, 7, ..., 255
       INZ
       BZE OUT     ; prime found if last divisor was 255
       INZ
       STZ TEMP
       LDA TRY
       DVZ
       BZN INLP    ; branch to new divisor if remainder non-zero
       DEC
       BNE GETPRM  ; branch to new try if divisor found besides try itself
OUT    RTS
 
; PRETST checks 2**P - 1 for a few small odd divisors first ...
; if definitely not prime, a non-zero Z is returned
 
PRETST ZER
       DEC
       JSR RIGHT
       STA LVAL    ; LVAL <-- 2**P - 1 ... possible Mersenne prime
       LZI 1
       STZ TEMP
PRELP  LDZ TEMP    ; loop tests division by 3, 5, 7, ..., 255
       INZ
       BZE PREOUT  ; Z = 0 exit indicates no small divisors found ...
       INZ
       STZ TEMP
       LDA LVAL
       DVZ
       BZN PRELP
       DEC
       BEQ PREOUT  ; Z = 0 exit here indicates LVAL is prime
       DEZ
PREOUT RTS
 
; LLTEST does Lucas-Lehmer test ... input is (HICNT,LOCNT), SHIFT, and pre-set
;   long integer size ... ACC returned as zero if Mersenne exponent
 
LLTEST LAI 4
LOOP   STA LVAL
       STA LDUP
       ZER
 
       MUL LVAL LDUP  ; time consuming step ... square LVAL
 
       EXA LVAL    ; next, decrement result twice ... result contained in
       BNE OK1     ;   ACC (low order) and LVAL (high order) after the
       EXA LVAL    ;   initial swap
       DEC
       EXA LVAL
OK1    DEC
       BNE OK2
       EXA LVAL
       DEC
       EXA LVAL
OK2    DEC
 
       JSR MODDIV  ; perform modulo division of result by 2**TRY - 1
 
       LDZ LOCNT   ; decrement the (HICNT,LOCNT) latch
       DEZ
       STZ LOCNT
       INZ
       BZN LOOP
       LDZ HICNT
       DEZ
       STZ HICNT
       INZ
       BZN LOOP
       RTS
 
; MODDIV uses shifts and additions to return remainder of (LVAL,ACC) divided
;   by 2**TRY - 1 in ACC ... SHIFT, with long integer size, implies TRY
 
MODDIV LDZ SHIFT   ; first, left shift SHIFT bits from low order result into
LP1    RLC         ;   high order result (initial carry irrelevant)
       EXA LVAL
       RLC
       EXA LVAL
       DEZ
       BZN LP1
 
       JSR RIGHT   ; next, restore low order result, with clear fill from left
 
       ADD LVAL    ; next, add high order result to low order result
 
       JSR LEFT    ; shift result left again to test for single excess bit
 
       BCC A1      ; save the bit temporarily
       INZ
A1     STZ TEMP
 
       JSR RIGHT   ; restore result with clear fill from left
 
       LDZ TEMP    ; add the excess bit to the result
       ADZ
 
       INC         ; finally, test if result exactly equals 2**TRY - 1
       JSR LEFT
       BEQ A2      ; if equal, do not restore
       JSR RIGHT
       DEC
A2     RTS
 
; RIGHT shifts ACC right SHIFT bits (with clear fill) and returns last bit
 
RIGHT  LDZ SHIFT
LP2    CLC
       RRC
       DEZ
       BZN LP2
       RTS
 
; LEFT shifts ACC left SHIFT bits (with clear fill) and returns last bit
 
LEFT   LDZ SHIFT
LP3    CLC
       RLC
       DEZ
       BZN LP3
       RTS
 
; SAVMER saves Mersenne exponent in high storage and bumps count
 
SAVMER LZI 6       ; set long integer size to 1.5 K
       LNH
       CLZ
       LNL
       LDA LAST    ; load last long variable, which is 1.5 K block above
       LDZ LOTRY   ;   the maximum 24 K used by the Lucas-Lehmer test,
       RRZ         ;   and rotate the current exponent into it
       LDZ HITRY
       RRZ
       STA LAST    ; save the data back in high storage
       LDZ MERCNT
       INZ         ; increment the current count of Mersenne primes found
       STZ MERCNT
       RTS
 
;****************************** END PROGRAM **********************************
