;           *********************************************************
;           *  Sample LIAL program to compute PI to high precision  *
;           *********************************************************
 
;********************************* REMARKS ***********************************
;
; This LIAL program computes PI by Machin's method, which uses the following:
;
;    1.  PI/4 = 4 * ARCTAN(1/5) - ARCTAN(1/239)
;
;    2.  ARCTAN(X) = X - (X**3)/3 + (X**5)/5 - (X**7)/7 + ...
;
; The number of bytes precision is set by N below.  5 * N bytes of long integer
; storage is required.  If there is a dynamic storage limit of 64 K imposed by
; the run module, then N is limited to about 11 K.  (The run module will need
; some dynamic storage, and the header and program must also be loaded into
; memory.)
;
; Program output:
;
;    1. First 2 * N bytes of storage contain packed-decimal fraction PI / 10,
;       with trailing insignificant digits included.
;
;    2. Next N bytes of storage contain binary PI / 4.  Some precision is lost
;       due to rounding in the division step and left shifting of the
;       intermediate ARCTAN(1/5) result.  The number of bits of precision
;       lost is less than two plus the binary log of N.  E.g., if N is 1024,
;       then at most 1.5 bytes of precision is lost in the 1024 byte result.
;
;    3. Bytes 19-20 and 21-22 of header array hold the summation indices of
;       the last terms for the power series of ARCTAN(1/239) and ARCTAN(1/5),
;       respectively, generated in the pi computation.  (Two times the index
;       plus one yields the inverse coefficient for the last term.)  This may
;       be used to get a precise error bound for the computation.
;
; Running time is roughly proportional to N-cube if the LIAL interpreter uses
; the classical order N-square division algorithm.
;
;************************** ASSEMBLER ASSIGNMENTS ****************************
 
; First, initialize some object file header values:
 
* N       = 64     ; number of bytes per long integer   \   ******************
* STORAGE = 320    ; need at least five times N          >  * MAY BE CHANGED *
* FLAGS   = $7     ; enables user break & bounds checks /   ******************
 
* PC      = BOOT   ; initialize program counter (default also ok here)
 
; Next, assign variable names to lowest available of the 16 long integers:
 
# POWER   = 0      ; holds odd powers of 5 or 239
# TERM    = 1      ; holds terms in power series sum
# SUM     = 2      ; holds arctangent power series sum
# DENOM   = 3      ; holds denominator prior to computation of term
 
; Following names are long integer aliases, since workspace can be reused:
 
# DECFRC  = 0      ; destination of all conversions (BDF, BDI, DBF, DBI)
# SAVDEC  = 1      ; temporary save area for result of first BDF call
 
; Next, assign variable names to Z-registers that are used:
 
# XINV    = 0      ; holds 5 or 239 (inverse of ARCTAN argument)
# SIGN    = 1      ; holds 0 or 255 (to flag addition or subtraction)
 
# HI      = 2      ; two-byte summation index for terms
# LO      = 3      ;   of ARCTAN power series
 
# SAVEHI  = 4      ; saves final summation index of
# SAVELO  = 5      ;   ARCTAN(1/5) computation
 
;****************************** PROGRAM **************************************
 
BOOT   JSR PIOVR4  ; returns PI / 4 in ACC ...
 
       LZI 5       ; change to PI / 10 by dividing by 5, then multiplying
       DVZ         ;   by 2, with an adjustment to least significant bit
       CLC         ;   to account for division remainder
       INZ
       SRZ
       SRZ
       BZE UP
       SEC
UP     RLC
 
       BDF         ; get leading packed-decimal fractional digits
       EXA DECFRC
       STA SAVDEC  ; save result of first conversion call
       LDA DECFRC
 
       BDF         ; get remaining packed-decimal fractional digits
       EXA SUM     ; get saved binary PI / 4 value
       EXA SAVDEC  ; swap with first conversion result
       END
       BRN BOOT    ; repeat program on restart
 
; PIOVR4 subroutine computes ACC <-- SUM <-- PI / 4.
 
PIOVR4 ZER         ; ACC <-- 0
       CLZ
       STZ SIGN    ; SIGN <-- 0 to flag addition
       LZI 5       ; Z <-- 5 for ARCTAN(1/5)
       JSR ARCTAN
 
       LDZ HI      ; save last ARCTAN(1/5) summation index ... last index
       STZ SAVEHI  ;   is one at which overflow occurs and division quotient
       LDZ LO      ;   would be zero ... there may be rounding contribution
       STZ SAVELO  ;   from term, however
 
       CLC         ; ACC <-- ACC * 4 with two shifts ... two bits of
       RLC         ;   precision is lost here
       RLC
       LZI 255
       STZ SIGN    ; SIGN <-- 255 to flag subtraction
       LZI 239     ; Z <-- 239 for ARCTAN(1/239) ... fall to ARCTAN
 
; ARCTAN subroutine computes ACC <-- ACC +/- ARCTAN(1/Z) depending on SIGN
 
ARCTAN STA SUM
       STZ XINV
       LAI 1
       STA POWER
       CLZ
       STZ HI      ; (HI,LO) <-- 0 ... two bytes are sufficient for the
       STZ LO      ;   summation index when N is less than 38 K
 
LOOP   LDZ XINV    ; POWER <-- POWER * XINV ... first time through, this
       MLZ         ; done only once, but same operation at end of loop
       BZN EXIT    ; insures that it is performed twice each time thereafter
       STA POWER
 
       LDZ LO      ; ACC <-- POWER * (HI,LO)
       MLZ
       BZN EXIT
       LDZ HI
       BZE AHEAD0
       STA DENOM   ; DENOM gets temporary use here when HI is non-zero
       LDA POWER
       CLZ
       RLZ         ; left shift ACC one byte
       BZN EXIT
       LDZ HI
       MLZ
       BZN EXIT
       ADD DENOM
       BVS EXIT
 
AHEAD0 CLC
       RLC         ; DENOM <-- 2 * ACC + POWER
       ADD POWER
       STA DENOM
 
       ZER         ; (ACC,TERM) <-- 1 ... numerator for division step
       STA TERM
       INC
 
       BCS DONE0   ; check for carry (from RLC) or overflow (from ADD) ...
       BVC AHEAD1  ;   if either found, then done, but will still have
       BRN DONE1   ;   rounding contribution from last term if not both
DONE0  BVS EXIT    ;   have occurred
DONE1  JSR UPSUM   ; ACC is 1 here for final contribution to SUM
       BRN EXIT
 
AHEAD1 DIV TERM DENOM   ; TERM <-- 1 / DENOM ... time-consuming step
 
       CLC         ; check whether twice remainder minus denominator is
       RLC         ;   positive ... if so, final carry is clear and TERM
       SUB DENOM   ;   must be incremented (rounding)
       RLC
 
 
       LDA TERM    ; SUM <-- SUM +/- (rounded TERM) depending on SIGN,
       BCS AHEAD2  ;   which flips
       INC
AHEAD2 JSR UPSUM
 
       LDZ LO      ; (HI,LO) <-- (HI,LO) + 1
       INZ
       STZ LO
       BZN AHEAD3
       LDZ HI
       INZ
       STZ HI
       BZN AHEAD3
       END         ; this abort only occurs if N larger than about 38 K
 
AHEAD3 LDA POWER   ; POWER <-- POWER * XINV
       LDZ XINV
       MLZ
       BZE LOOP
 
EXIT   LDA SUM     ; exits occur whenever DENOM would become too large
       RTS
 
; UPSUM subroutine sets SUM <-- SUM +/- ACC depending on SIGN, which flips
 
UPSUM  LDZ SIGN
       BZE AHEAD4
       NEG
       INZ
       INZ
AHEAD4 DEZ
       STZ SIGN
       ADD SUM
       STA SUM
       RTS
 
;****************************** END PROGRAM **********************************
