;    **********************************************************************
;    *  Sample LIAL program to find large primes by probabilistic method  *
;    **********************************************************************
 
;********************************* REMARKS ***********************************
;
; This LIAL program uses a probabilistic test due to M. Rabin to find large
; primes.  The test takes as input an odd integer p greater than 1 and a
; random integer in the range [2,p-1].  The test outputs the result that
; 'p is not prime' or 'p is prime with probability at least 75%'.  If the
; test is run n times (with different random input, but with the same odd
; prime candidate p) and each time the result is 'probably prime', then the
; probability that p is prime is at least 1 - 1/4**n.  For n = 5, this amounts
; to a probability of better than 99.9%.
;
; Why use a probabilistic test?  Speed.  There is no sure-fire test that can
; find arbitrary large primes in a reasonable amount of time, particularly on
; microcomputers.  There are some fast tests that find special types of large
; primes, e.g., the Lucas-Lehmer test for finding Mersenne primes.  But a
; reasonable way to determine if an arbitrary large integer is prime (even
; though not with 100% certainty) is useful, e.g., in cryptographic
; applications.
;
; The program here takes as input an N-byte starting point (the value for N
; and the actual starting point may be modified below).  The first prime
; greater than the starting point is the program output.  Successive program
; restarts will find successively larger primes.  The number of Rabin test
; iterations (and thus the desired probability level) may also be modified
; below, if desired.
;
; To optimize program speed, some trial divisions by small odd integers are
; performed before resorting to the Rabin test.
;
;************************** ASSEMBLER ASSIGNMENTS ****************************
 
; First, initialize some object file header values:
 
* PC      = BOOT   ; initialize program counter (default ok)
* SEED    = 987654 ; set 4-byte SEED to any non-zero value (default of 1 ok)
 
* N       = 16     ; number of bytes per long integer   \   ******************
* STORAGE = 112    ; need at least seven times N         >  * MAY BE CHANGED *
* FLAGS   = $7     ; enables user break & bounds checks /   ******************
 
; Next, initialize the N-byte starting point, which otherwise defaults to zero
 
% $FE $DC $BA $98 $76 $54 $32 $10 $01 $23 $45 $67 $89 $AB $CD $F5
 
; Next, assign constant value for the desired number of Rabin test iterations:
 
# RABITER = 5      ; allowed range is 0 - 255 (0 = 256)   <-- MAY BE CHANGED
 
; Next, assign variable names to lowest available of the 16 long integers:
 
# RESULT  = 0      ; holds MODEXP and RABTST results
# DIVSR   = 0      ; holds small odd trial divisors (alias)
# TEMP    = 1      ; holds temporary MUL or DIV operand
# COUNT   = 2      ; holds shift counter for MODEXP and RABTST
# ODDTRY  = 3      ; holds odd values for prime candidates
# POWER   = 4      ; holds exponent for MODEXP and RABTST
# RNDVAL  = 5      ; holds the random values for RABTST
 
; Next, assign variable name for auxilliary z-register used
 
# ZHOLD   = 0      ; holds odd trial divisors from 3 - 255
 
;****************************** PROGRAM **************************************
 
BOOT   RRC         ; insure that starting point is odd
       SEC
       RLC
       STA ODDTRY
NEXT   JSR GETPRM  ; find first prime greater than ODDTRY
       LDA ODDTRY
       BDI         ; convert most of result to packed-decimal digits
       EXA 0       ; move low-order packed-decimal digits to
       STA 1       ;   second long integer area
       LDA 0       ; load unconverted part of result for next conversion
       BDI         ; convert rest of result in first long integer area
       LDA ODDTRY  ; put hex result at start of long integer storage
       END
       BRN NEXT    ; restart finds next prime...
 
; GETPRM routine inputs ODDTRY and outputs next larger prime in ODDTRY
;    --- note four separate routine exits (RTS)
 
GETPRM LDA ODDTRY ; first increment ODDTRY (usually by two)
       INC
       BNE A0     ; equality occurs only on wrap-around to small integers
       INC
       INC
A0     INC
       STA ODDTRY
 
       LZI 1      ; next, do trial divisions by 3, 5, 7, 9, ..., 255
       STZ ZHOLD
LOOP0  LDZ ZHOLD
       INZ
       BZE A1     ; quit when 255 bumped to 0
       INZ
       STZ ZHOLD
       LDA ODDTRY
       DVZ        ; fast one-byte division
       BZN LOOP0  ; loop if non-zero remainder
       DEC
       BNE GETPRM ; try next ODDTRY if quotient > 1
       RTS        ; otherwise ODDTRY is prime < 255
 
A1     LDA ODDTRY ; ODDTRY is prime at this point if it is a 16-bit
       LEN        ;   long number or less
       LZI 17
       SBZ
       BPL A2     ; branch if 17 or more bits in ODDTRY
       RTS
 
A2     LAI 255    ; next, do trial divisions by small odd divisors
       STA DIVSR  ;   larger than 255 but less than roughly binary
LOOP1  LDA DIVSR  ;   logarithm of ODDTRY ... slower divisions
       INC
       INC
       STA DIVSR  ; bump small divisor by two
 
       LDA ODDTRY
       STA TEMP
       LEN        ; ACC <-- roughly binary logarithm of ODDTRY
       SUB DIVSR
       BMI A3     ; quit trial divisions if divisor too large
 
       ZER
       DIV TEMP DIVSR
       BNE LOOP1  ; try next divisor, if non-zero remainder
       LDA TEMP
       DEC
       BNE GETPRM ; if quotient > 1, divisor found, so try next ODDTRY
       RTS        ; otherwise, ODDTRY is a small prime > 255
 
A3     LZI RABITER
LOOP2  JSR RABTST ; run Rabin test ...
       BNE GETPRM ; if ACC non-zero, ODDTRY not prime, so try next ODDTRY
       DEZ
       BZN LOOP2
       RTS        ; here, ODDTRY is probably prime
 
; RABTST routine takes as input ODDTRY (and system SEED) and returns
;   ACC <-- 1 if not prime, else ACC <-- 0 if 75% probable prime
 
RABTST LDA ODDTRY ; first, set RNDVAL in range [2,ODDTRY-1]
       DEC
       DEC
       DEC
       RND
       INC
       INC
       STA RNDVAL
 
       LDA ODDTRY ; next, set POWER
       CLC
       RRC
       CLC
LOOP3  RRC
       BCC LOOP3
       RLC        ; restores low order set bit
       STA POWER
 
       JSR MODEXP ; RESULT <-- RNDVAL ** POWER modulo ODDTRY
       LDA RESULT
       DEC
       BEQ OUT    ; probable prime if result was 1
 
       LDA ODDTRY
       RRC        ; (ignore incoming garbage bit)
       STA COUNT  ; sets shift counter for next loop
 
LOOP4  LDA RESULT
       STA TEMP
       INC
       SUB ODDTRY
       BEQ OUT    ; probable prime if RESULT was ODDTRY - 1
 
       LDA COUNT
       RRC        ; (ignore incoming garbage bit)
       STA COUNT
       BCS A4     ; done looping, and not prime, if carry set
 
       ZER
       MUL RESULT TEMP
       DIV RESULT ODDTRY
       STA RESULT
       DEC
       BNE LOOP4  ; if RESULT was 1, then not prime, else loop back
 
A4     LAI 1
OUT    RTS
 
; MODEXP routine computes RESULT <-- RNDVAL ** POWER modulo ODDTRY by
;   method of exponentiation by repeated squaring and multiplication
 
MODEXP LAI 1      ; initialize COUNT and RESULT to 1
       STA RESULT
 
LOOP5  STA COUNT
       LDA RESULT
       STA TEMP   ; TEMP duplicates RESULT for square operation
       DEC
       BEQ A5     ; skip squaring step if RESULT still 1
 
       ZER
       MUL RESULT TEMP
       DIV RESULT ODDTRY
       STA RESULT ; RESULT <-- remainder
 
A5     LDA POWER
       RLC        ; (ignore incoming garbage bit at right)
       STA POWER
       BCC A6     ; skip multiplication step if next POWER bit clear
 
       ZER
       MUL RESULT RNDVAL
       DIV RESULT ODDTRY
       STA RESULT ; RESULT <-- remainder
 
A6     LDA COUNT  ; shift COUNT to look for initial set bit
       RLC        ; (ignore incoming garbage bit at right)
       BCC LOOP5
       RTS
 
;**************************** END PROGRAM **********************************
